;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "helm-descbinds" "20250705.942"
  "A convenient `describe-bindings' with `helm'."
  '((helm "1.5"))
  :url "https://github.com/emacs-helm/helm-descbinds"
  :commit "0aff44badad976ebf2666a7e9b6ddf4db53e59e5"
  :revdesc "0aff44badad9"
  :keywords '("helm" "help")
  :authors '(("Taiki SUGAWARA" . "buzz.taiki@gmail.com"))
  :maintainers '(("Taiki SUGAWARA" . "buzz.taiki@gmail.com")))
